import unittest
from click.testing import CliRunner
from ..create_cmd import create_cmd
from pathlib import Path
import tempfile
import shutil
import os


class TestCreate(unittest.TestCase):

    def setUp(self):
        self.test_dir = Path(tempfile.mkdtemp())
        os.chdir(self.test_dir)
        # print(f"test_dir='{self.test_dir}'")

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_basic_cmd(self):
        runner = CliRunner()
        result = runner.invoke(create_cmd, ['--template',
                                                    'basic',
                                                    '--network',
                                                    'testnet',
                                                    'test-app'])
        assert result.exit_code == 0
        # assert result.output == "Created cardano defaults from 'basic' template for network 'testnet': 'test-app'\n"

        # # Create a file in the temporary directory
        # f = open(path.join(self.test_dir, 'test.txt'), 'w')
        # # Write something to it
        # f.write('The owls are not what they seem')
        # # Reopen the file and check if what we read back is the same
        # f = open(path.join(self.test_dir, 'test.txt'))
        # self.assertEqual(f.read(), 'The owls are not what they seem')


import click

from .query import query_cmd
from .address_cmd import address_cmd
from .stake_address_cmd import stake_address_cmd


@click.group("node")
def node_cmd():
    """Cardano Node CLI command"""
    pass


# TODO:bfloyd remove later before 1.0
@click.command("cli")
@click.pass_context
def cli_cmd(ctx):
    ctx.fail("cli command group is obsolete. please use `node` instead for the same features. e.g., `cardanopy node query tip`")


# export
node_cmd.add_command(query_cmd)
node_cmd.add_command(address_cmd)
node_cmd.add_command(stake_address_cmd)



import os
from pathlib import Path
from ..core.cardanopy_error import CardanoPyError


class CardanoNode(object):

    @staticmethod
    def validate_environment():
        cardano_node_socket_path = os.getenv('CARDANO_NODE_SOCKET_PATH')
        if not cardano_node_socket_path:
            raise CardanoPyError(f"Invalid CARDANO_NODE_SOCKET_PATH. Undefined.", 1)

        cardano_node_socket_path = Path(cardano_node_socket_path)
        if not cardano_node_socket_path.exists():
            raise CardanoPyError(f"Invalid CARDANO_NODE_SOCKET_PATH. '{cardano_node_socket_path}' does not exist.", 1)

        CardanoNode.get_network()

    @staticmethod
    def get_network():
        cardano_network = os.getenv('CARDANO_NETWORK')
        if not cardano_network:
            raise CardanoPyError(f"Invalid CARDANO_NETWORK. Undefined.", 1)

        return cardano_network


import click
from .node.cardano_cli import CardanoCli
from .node.cardano_node import CardanoNode


@click.command("network")
@click.option('-p', '--param', 'param', is_flag=True, help="print the network param help for cardano-cli")
def network_cmd(param):
    """Get the cardano network"""

    if not param:
        print(CardanoNode.get_cli_network())
    else:
        print(" ".join(CardanoCli.get_network_cli_args()))


# import click
# from .core.cardanopy_config import CardanoPyConfig
#
#
# @click.command()
# @click.argument('property_name', type=str, required=True)
# @click.argument('property_value', type=str)
# @click.argument('target_config_dir', type=str)
# @click.pass_context
# def config(ctx, property_name, property_value, target_config_dir):
#     """Config command"""
#
#     cardanopy_config = CardanoPyConfig()
#     try:
#         cardanopy_config.load(target_config_dir)
#     except ValueError as e:
#         ctx.fail(e.args)
#         return 1
#
#     if not property_value:
#         print(cardanopy_config.get(property_name))
#     else:
#         cardanopy_config.set(property_name, property_value)
#
#         try:
#             cardanopy_config.save(target_config_dir)
#         except ValueError as e:
#             ctx.fail(e.args)
#             return 1



import unittest
from click.testing import CliRunner
from ..create_cmd import create_cmd
from ..run_cmd import run_cmd
from pathlib import Path
import tempfile
import shutil
import os


class TestRun(unittest.TestCase):

    def setUp(self):
        self.test_dir = Path(tempfile.mkdtemp())
        os.chdir(self.test_dir)
        # print(f"test_dir='{self.test_dir}'")

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_run_cmd(self):
        app_dir = str(self.test_dir.joinpath('test-app'))
        runner = CliRunner()
        create_result = runner.invoke(create_cmd, ['--template',
                                                'bp-k8s',
                                                '--network',
                                                'testnet',
                                                'test-app'])
        # print(create_result.output)
        assert create_result.exit_code == 0
        run_result = runner.invoke(run_cmd, ['--dry-run', '--block-producer', app_dir])
        # print(generate_result.output)
        assert run_result.exit_code == 0
        # assert "docker run --name basic -d --env CARDANO_NODE_SOCKET_PATH=/app/node.socket --env CARDANO_NETWORK=testnet -p 3001:3001" in run_result.output
        # #  -v /private/var/folders/dj/qy6c5r217zxf4xf2rspds_km0000gn/T/tmpf9759ir0/test-app:/app
        # assert "floydcraft/cardano-py-slim:latest run /app" in run_result.output
        # assert "docker exec -it basic bin/bash" in run_result.output


import click
from ..core.cardanopy_config import CardanoPyConfig
from .docker_helper import DockerHelper


@click.command("run")
@click.option('-p', '--pull', 'pull', is_flag=True, help="pull the docker image. Instead of using local docker image cache")
@click.option('-s', '--stop', 'stop', is_flag=True, help="stop and remove the docker image before running")
@click.option('-b', '--bash', 'bash', is_flag=True, help="override docker image entrypoint with /bin/bash")
@click.option('-d', '--dry-run', 'dry_run', is_flag=True, help="print the mutable commands")
@click.option('-s', '--sub', 'subs', multiple=True, type=str, default=tuple(), help="Substitutions for configs")
@click.argument('target_config_dir_or_file', type=str)
@click.pass_context
def run_cmd(ctx, pull, stop, bash, dry_run, subs, target_config_dir_or_file):
    """Docker Run helper command"""

    try:
        target_config_dir = CardanoPyConfig.try_get_valid_config_dir(target_config_dir_or_file)
        target_config_file = CardanoPyConfig.try_get_valid_config_file(target_config_dir_or_file)

        cardanopy_config = CardanoPyConfig()
        cardanopy_config.load(target_config_file, subs)

        if pull:
            DockerHelper.pull_container(cardanopy_config.docker.image, dry_run)

        is_container_running = DockerHelper.is_container_running(cardanopy_config.docker.name, dry_run)

        if is_container_running:
            if stop:
                DockerHelper.stop_container(cardanopy_config.docker.name, dry_run)
            else:
                DockerHelper.exec_bash(cardanopy_config.docker.name, target_config_dir, dry_run)
                return 0

        if DockerHelper.is_container_exited(cardanopy_config.docker.name, dry_run):
            DockerHelper.remove_container(cardanopy_config.docker.name, dry_run)

        DockerHelper.run_cardano_node(container_name=cardanopy_config.docker.name,
                                      target_config_dir=target_config_dir,
                                      socket_path=cardanopy_config.socketPath,
                                      network=cardanopy_config.network,
                                      port=cardanopy_config.port,
                                      docker_root_volume=cardanopy_config.docker.rootVolume,
                                      docker_image=cardanopy_config.docker.image,
                                      docker_mount=cardanopy_config.docker.mount,
                                      bash=bash,
                                      dry_run=dry_run)

        if not bash:
            DockerHelper.exec_bash(cardanopy_config.docker.name, target_config_dir, dry_run)
    except Exception as ex:
        ctx.fail(f"docker_cmd:run_cmd(pull={pull}, "
                 f"dry_run={dry_run}, "
                 f"target_config_dir_or_file='{target_config_dir_or_file}') failed: {type(ex).__name__} {ex.args}")
        return 1



__version__ = "0.1.8"
__license__ = "Apache-2.0 License"
__title__ = "cardanopy"

import click

from .run_cmd import run_cmd
from .create_cmd import create_cmd
from .node import node_cmd, cli_cmd
from .docker_cmd import docker_cmd
from .k8s_cmd import k8s_cmd
from .generate_cmd import generate_cmd
from .network_cmd import network_cmd
from .healthcheck_cmd import healthcheck_cmd

@click.group()
@click.version_option(version=__version__)
@click.pass_context
def main(ctx):
    pass


# export
main.add_command(run_cmd)
main.add_command(create_cmd)
main.add_command(node_cmd)
main.add_command(cli_cmd)
main.add_command(docker_cmd)
main.add_command(k8s_cmd)
main.add_command(generate_cmd)
main.add_command(network_cmd)
main.add_command(healthcheck_cmd)
# disable config cli for now (might remove)
# main.add_command(config)



import click
from .cardano_cli import CardanoCli
from ..core.cardanopy_error import CardanoPyError


@click.group("query")
@click.pass_context
def query_cmd(ctx):
    """Node query commands. Will query the local node"""


@query_cmd.command("tip")
@click.option('-d', '--dry-run', 'dry_run', is_flag=True, help="print the mutable commands")
@click.pass_context
def tip_cmd(ctx, dry_run):
    """Get the node's current tip (slot no, hash, block no)."""
    try:
        print(query_tip(dry_run), end="")
    except CardanoPyError as cpe:
        ctx.fail(cpe.message)
        return cpe.return_code


def query_tip(dry_run) -> str:
    """Get the node's current tip (slot no, hash, block no)."""
    result = CardanoCli.execute(cmd=["cardano-cli", "query", "tip"], dry_run=dry_run, include_network=True)
    if result.stderr_str:
        raise CardanoPyError(result.stderr_str)
    elif not result.stdout_str:
        raise CardanoPyError("null result for query tip")
    else:
        return result.stdout_str


@query_cmd.command("utxo")
@click.option('-d', '--dry-run', 'dry_run', is_flag=True, help="print the mutable commands")
@click.pass_context
def utxo_cmd(ctx, dry_run):
    """Get the node's current UTxO with the option of filtering by address(es)"""

    try:
        CardanoCli.execute(cmd=["cardano-cli", "query", "utxo"], dry_run=dry_run, include_network=True)
    except CardanoPyError as cpe:
        ctx.fail(cpe.message)
        return cpe.return_code



import click
import subprocess
import os


@click.group("address")
@click.pass_context
def address_cmd(ctx):
    """Payment address commands"""


@address_cmd.command("key-gen")
@click.option('--normal-key', 'key_type', flag_value='normal-key', default=True, help='Use a normal Shelley-era key (default).')
@click.option('--extended-key', 'key_type', flag_value='extended-key', help='Use an extended ed25519 Shelley-era key.')
@click.option('--byron-key', 'key_type', flag_value='byron-key', help='Use a Byron-era key.')
@click.option('--verification-key-file', 'verification_key_file', required=True, type=str, help='Output filepath of the verification key.')
@click.option('--signing-key-file', 'signing_key_file', required=True, type=str, help='Output filepath of the signing key.')
@click.pass_context
def key_gen_cmd(ctx, key_type, verification_key_file, signing_key_file):
    """Create an address key pair."""

    try:
        subprocess.run(["cardano-cli",
                        "address",
                        "key-gen",
                        f"--{key_type}",
                        "--verification-key-file", verification_key_file,
                        "--signing-key-file", signing_key_file
                        ])
    except Exception as ex:
        ctx.fail(f"cli:address_cmd:key_gen_cmd(key_type={key_type}, "
                 f"verification_key_file={verification_key_file}, "
                 f"signing_key_file='{signing_key_file}') failed: {type(ex).__name__} {ex.args}")
        return 1


@address_cmd.command("build")
@click.pass_context
@click.option('--payment-verification-key-file', 'payment_verification_key_file', required=True, type=str, help='Output filepath of the verification key.')
@click.option('--stake-verification-key-file', 'stake_verification_key_file', required=True, type=str, help='Output filepath of the signing key.')
@click.option('--out-file', 'out_file', type=str, help='Optional output file. Default is to write to stdout')
@click.pass_context
def build_cmd(ctx, payment_verification_key_file, stake_verification_key_file, out_file):
    """Create an address key pair."""

    try:
        cardano_network = os.getenv('CARDANO_NETWORK')
        if cardano_network == "mainnet":
            cardano_network_params = ['--mainnet']
        elif cardano_network == "testnet":
            cardano_network_params = ['--testnet-magic', '1097911063']
        else:
            ctx.fail("missing or invalid param: mainnet or testnet")
            return

        subprocess.run(["cardano-cli",
                        "address",
                        "build",
                        "--payment-verification-key-file", payment_verification_key_file,
                        "--stake-verification-key-file", stake_verification_key_file]
                        + ([] if out_file is None else ['--out-file', out_file])
                        + cardano_network_params)
    except Exception as ex:
        ctx.fail(f"cli:address_cmd:build_cmd(payment_verification_key_file={payment_verification_key_file}, "
                 f"stake_verification_key_file={stake_verification_key_file}, "
                 f"out_file='{out_file}') failed: {type(ex).__name__} {ex.args}")
        return 1



from pathlib import Path
from .cardanopy_config import CardanoPyConfig
import json
import os


class Substitution(object):

    @staticmethod
    def generate(dry_run: bool, target_config_dir: Path, config: CardanoPyConfig, subs: tuple = tuple()):
        for dir_name, subdirList, fileList in os.walk(target_config_dir):
            dir_path = Path(dir_name)
            for file_name in fileList:
                file_path = dir_path.joinpath(file_name)
                if ".template" in file_name:
                    output_config_file = dir_path.joinpath(file_name.replace(".template", ""))
                    if dry_run:
                        print(f"generate config '{output_config_file}' from '{file_path}'")
                    else:
                        try:
                            with open(file_path, "r") as file:
                                config_data_str = file.read()
                        except Exception as ex:
                            raise ValueError(
                                f"Failed to load template. '{file_path}'. {type(ex).__name__} {ex.args}")

                        try:
                            config_data_resolved_str = config.substitute(config_data_str, subs)
                        except Exception as ex:
                            raise ValueError(
                                f"Failed to substitute template. '{file_path}'. {type(ex).__name__} {ex.args}")

                        try:
                            with open(output_config_file, "w") as file:
                                config_local = json.loads(config_data_resolved_str)
                                print(json.dumps(config_local, indent=4), file=file)
                        except Exception as ex:
                            raise ValueError(
                                f"Failed to save generated config. '{output_config_file}'. {type(ex).__name__} {ex.args}")

                        print(f"generated config '{output_config_file}'")
